#!/usr/bin/env pythoon3
"""Read AVS dats file."""
import sys
import re
from argparse import ArgumentParser
from pathlib import Path

import dateutil.parser
import openpyxl as XL
from openpyxl.cell.cell import MergedCell
from openpyxl.utils.exceptions import InvalidFileException

try:
    import itkdb_gtk

except ImportError:
    cwd = Path(sys.argv[0]).parent.parent
    sys.path.append(cwd.as_posix())

from itkdb_gtk import ITkDBlogin, ITkDButils

sk_defaults = {
    "institution": "AVS",
    "runNumber": "1",
}


class AVSDataException(Exception):
    """AVSData exception class."""

    def __init__(self, message):
        """Call the base class constructor with the parameters it needs."""
        super().__init__(message)


def create_weight(session, SN, the_date=None, manager="", passed=True, problems=False, comments=None):
    """Creates the dictionary for a WEIGHT test.

    Args:
        session: the DB session
        SN: Serial Number
        the_date: the date of the test
        manager: manager name
        passed: if test passed or not
        problems: if problems were found during test
        comments: list of comments to append to the test

    """
    if comments is None:
        comments = []

    the_date = ITkDButils.get_db_date(the_date)
    out = ITkDButils.get_test_skeleton(session, "CORE_PETAL", "WEIGHING", sk_defaults)
    out['component'] = SN
    out['date'] = the_date
    out["institution"] = "AVS"
    out["runNumber"] = "1"
    out['passed'] = passed
    out['problems'] = problems
    out['properties']['PRODUCTION_MANAGER'] = manager
    out['comments'] = comments
    return out


def create_manufacturing(session, SN, the_date=None, manager="", passed=True, problems=False, comments=None):
    """Create the dictionary or the MANUFACTURING test.

    Args:
        session: the DB session
        SN: Serial Number
        the_date: the date of the test
        manager: manager name
        passed: if test passed or not
        problems: if problems were found during test
        comments: list of comments to append to the test

    """
    if comments is None:
        comments = []
    the_date = ITkDButils.get_db_date(the_date)
    out = ITkDButils.get_test_skeleton(session, "CORE_PETAL", "MANUFACTURING", sk_defaults)
    out['component'] = SN
    out['date'] = the_date
    out["institution"] = "AVS"
    out["runNumber"] = "1"
    out['passed'] = passed
    out['problems'] = problems
    out['properties']['PRODUCTION_MANAGER'] = manager
    out['comments'] = comments

    return out


def create_visual_inpection(session, SN, the_date=None, operator="", passed=True, problems=False, comments=None):
    """Create Visual Inspection test skeleton."""
    if comments is None:
        comments = []
    the_date = ITkDButils.get_db_date(the_date)
    out = ITkDButils.get_test_skeleton(session, "CORE_PETAL", "VISUAL_INSPECTION", sk_defaults)
    out['component'] = SN
    out["institution"] = "AVS"
    out["runNumber"] = "1"
    out['date'] = the_date
    out['passed'] = passed
    out['problems'] = problems
    out['properties']['OPERATOR'] = operator
    out['comments'] = comments

    return out


def create_delamination_test(session, SN, the_date=None, operator="", passed=True, problems=False, comments=None):
    """Create the delamination test JSON.

    Args:
        session: the DB session
        SN: Serial Number
        the_date: the date of the test
        operator: operator name
        passed: if test passed or not
        problems: if problems were found during test
        comments: list of comments to append to the test

    """
    if comments is None:
        comments = []

    the_date = ITkDButils.get_db_date(the_date)
    out = ITkDButils.get_test_skeleton(session, "CORE_PETAL", "DELAMINATION", sk_defaults, {"boolean": False})
    out['component'] = SN
    out['date'] = the_date
    out["institution"] = "AVS"
    out["runNumber"] = "1"
    out['passed'] = passed
    out['problems'] = problems
    out['properties']['OPERATOR'] = operator
    out['comments'] = comments

    return out


def create_grounding_test(session, SN, the_date=None, operator="", passed=True, problems=False, comments=None):
    """Create grounding test.

    Args:
        session: the DB session
        SN: Serial Number
        the_date: the date of the test
        operator: operator name
        passed: if test passed or not
        problems: if problems were found during test
        comments: list of comments to append to the test

    """
    if comments is None:
        comments = []
    the_date = ITkDButils.get_db_date(the_date)
    out = ITkDButils.get_test_skeleton(session, "CORE_PETAL", "GROUNDING_CHECK", sk_defaults, {"boolean": False})
    out['component'] = SN
    out['date'] = the_date
    out["institution"] = "AVS"
    out["runNumber"] = "1"
    out['passed'] = passed
    out['problems'] = problems
    out['properties']['OPERATOR'] = operator
    out['comments'] = comments

    return out


def create_metrology_test(session, SN, the_date=None, operator="", passed=True, problems=False, comments=None):
    """Metrology test.

    Args:
        session: the DB session
        SN: Serial Number
        the_date: the date of the test
        operator: operator name
        passed: if test passed or not
        problems: if problems were found during test
        comments: list of comments to append to the test

    """
    if comments is None:
        comments = []

    the_date = ITkDButils.get_db_date(the_date)
    out = ITkDButils.get_test_skeleton(session, "CORE_PETAL", "METROLOGY_AVS",
                                       sk_defaults, {"integer": -1, "float": -1.0})
    out['component'] = SN
    out['date'] = the_date
    out['passed'] = passed
    out["institution"] = "AVS"
    out["runNumber"] = "1"
    out['problems'] = problems
    out['properties']['OPERATOR'] = operator
    out['comments'] = comments

    return out


def split_comp_list(lst):
    """Split a  list of components separated by  various possible characters."""
    if lst is None:
        return []

    if isinstance(lst, float):
        return [lst]

    if isinstance(lst, int):
        return [lst]

    out = [lst]
    for sep in ['/', '\\', '\n']:
        if lst.find(sep) >= 0:
            out = [x.strip() for x in lst.split(sep)]
            break

    return out


def get_comments(txt):
    """Return test DB comment."""
    return split_comp_list(txt)


def get_float(cell, separator=None, default=0.0):
    """Return float from string."""
    txt = cell.value
    if txt is None:
        return default

    if separator is None:
        if isinstance(txt, float):
            return txt

        if isinstance(txt, int):
            return float(txt)

        if isinstance(txt, str):
            try:
                txt = txt.replace(',', '.')
                rr = re.findall(r"[-+]?[.]?[\d]+(?:,\d\d\d)*[\.]?\d*(?:[eE][-+]?\d+)?", txt)
                if len(rr) == 0:
                    return default
                
                val = float(rr[0])
                return val
            except ValueError:
                return default

        return default

    else:
        values = []
        for val in split_comp_list(txt):
            if isinstance(val, float) or isinstance(val, int):
                values.append(val)
            else:
                try:
                    v = float(val.strip().replace(',', '.'))
                except ValueError:
                    print("get_float: Cannot convert {} in {}".format(val, cell.coordinate))
                    v = default

                values.append(v)

        return values

def get_int(cell, default=None):
    """Get an int from a cell."""
    value = cell.value
    if value is None:
        return default

    return int(value)

def get_boolean(cell):
    """Get a boolean from a cell."""
    value = cell.value
    if value is None:
        return False

    else:
        txt = value.strip().lower()
        return txt == "pass"


def get_text(cell):
    """Get a string from a cell."""
    value = cell.value
    if value:
        value = value.strip()

    return value


def get_res_and_accep(sheet, indx):
    """Return result and acceptancee."""
    sval = sheet["g{}".format(indx)].value
    if isinstance(sval, float) or isinstance(sval, int):
        val = sval
        
    else:
        sval = ' '.join(sval.strip().split()).split()
        
        scale = 1.0
        try:
            rr = re.findall(r"[-+]?[.]?[\d]+(?:,\d\d\d)*[\.]?\d*(?:[eE][-+]?\d+)?", sval[0])
            if len(rr) == 0:
                val = 0.0
            else:
                val = float(rr[0])
                if len(sval)>1:
                    U = sval[1].upper()[0]
                    if U=='G':
                        scale = 1000
                    
        except ValueError:
            val = 0
        
        val = val * scale
    #val = get_float(sheet["g{}".format(indx)])
    pass_val = sheet["h{}".format(indx)]
    if pass_val.value is None: 
        # Operator did not set the PASS/FAIL thing
        accept = True
    else:
        accept = get_boolean(pass_val)
        
    return val, accept


def find_idx(lst, val):
    """Return index of occurence of val in lst."""
    R = re.compile(val, re.DOTALL|re.IGNORECASE)
    idx = [i for i, x in enumerate(lst) if (x and R.search(x))]
    return idx


def cell_value(sheet, coord):
    """Return the cell value."""
    cell = sheet[coord]
    if not isinstance(cell, MergedCell):
        return cell.value

    # "Oh no, the cell is merged!"
    for cell_range in sheet.merged_cells.ranges:
        if coord in cell_range:
            return cell_range.start_cell.value

    return None


def check_for_problems(sheet, the_test, row_range):
    """Finds FAIL massages in the Acceptance column."""
    nfail = 0
    for row in range(row_range[0], row_range[1]):
        txt = get_text(sheet["h{}".format(row)])
        if txt is None:
            continue

        txt = txt.lower()
        if txt[0] == 'f':
            nfail += 1
            hdr = get_text(sheet["d{}".format(row)])
            reason = cell_value(sheet, "i{}".format(row))

            if reason:
                if len(reason) < 20:
                    msg = "{}: {}".format(hdr, reason)
                    the_test["defects"].append({"name": msg})
                else:
                    the_test["defects"].append({"name": hdr,
                                                "description": reason})

    if nfail:
        the_test["passed"] = False


def get_coreID(bustapeID):
    """Build core SN from bus tape SN."""
    SN = "20USEBC" + bustapeID[-7:]
    return SN


def readFATfile(session, file_path, SN=None):
    """Read data from FAT excel file.

    Args:
        session: the DB session
        file_path: File path
        SN: COre serial number

    """
    # Open spreadsheet
    try:
        wb = XL.load_workbook(file_path)
    except InvalidFileException as ee:
        print("Could not open input file: ", file_path)
        print(ee)
        return None

    # Assume active sheet is the good one, otherwise will have to find in wb.sheetnames
    sheet = wb.active
    if sheet.max_row < 50 or sheet.max_column < 9:
        raise AVSDataException("Wrong FAT file")

    # Check the SN of the petal core
    if SN is None or len(SN) == 0:
        coreID = split_comp_list(sheet['C6'].value)
        if len(coreID) == 0:
            raise AVSDataException("Cannot figure out core SN in FAT file.")

        for cID in coreID:
            cmp = ITkDButils.get_DB_component(session, cID)
            if cmp["type"]["code"] == "BT_PETAL_FRONT":
                SN = get_coreID(cID)
                break

    batch = sheet['E6'].value
    operator = sheet['G6'].value

    txt = list(map(str.strip, sheet['i4'].value.split(':')))[1]
    test_date = dateutil.parser.parse(txt)

    # Get the test index
    test_name = [str(sheet[x][1].value) for x in range(1, sheet.max_row)]
    tests = [str(sheet[x][3].value) for x in range(1, sheet.max_row)]

    # This is to avoid adding 1 for cell names...
    tests.insert(0, None)
    test_name.insert(0, None)

    #
    # Visual inspection
    vi_text = get_text(sheet['i9'])
    vi_result = get_text(sheet['g9'])
    vi_pass = sheet['h9'].value.strip().lower() == "pass"
    vi_defects = []
    if vi_pass:
        if vi_result and len(vi_result):
            if vi_text and len(vi_text):
                vi_text = vi_result + '\n' + vi_text
            else:
                vi_text = vi_result

    else:
        vi_defects.append({"name": "PETAL_VI_DEFECT", "description": vi_result})

    vi_test = create_visual_inpection(session, SN, test_date, operator, vi_pass,
                                      comments=get_comments(vi_text))
    for df in vi_defects:
        vi_test["defects"].append(df)

    #
    # Delamination test
    dl_text = get_text(sheet['i10'])
    dl_result = get_text(sheet['g10'])
    dl_pass = sheet['h10'].value.strip().lower() == "pass"
    dl_defects = []
    if dl_pass:
        if dl_result and len(dl_result):
            if dl_text and len(dl_text):
                dl_text = dl_result + '\n' + dl_text
            else:
                dl_text = dl_result

    else:
        dl_defects.append({"name": "PETAL_DL_DEFECT",
                           "description": dl_result})

    delamination_test = create_delamination_test(session, SN, test_date, operator, dl_pass,
                                                 comments=get_comments(dl_text))
    for df in dl_defects:
        delamination_test["defects"].append(df)

    #
    # Conductivity
    # TODO: read proper rows
    grounding_test = create_grounding_test(session, SN, test_date, operator)
    cond_val, cond_pass = get_res_and_accep(sheet, tests.index("COND"))
    if "INS_LOOP" in tests:
        loop_val, loop_pass = get_res_and_accep(sheet, tests.index("INS_LOOP"))
    else:
        loop_val, loop_pass = get_res_and_accep(sheet, tests.index("INS"))

    if "INS_LOOP_GND" in tests:
        loop_gnd_val, loop_gnd_pass = get_res_and_accep(sheet, tests.index("INS_LOOP_GND"))
    else:
        loop_gnd_val, loop_gnd_pass = get_res_and_accep(sheet, tests.index("INS_FACE"))

    passed = cond_pass and loop_pass and loop_gnd_pass
    grounding_test["passed"] = passed
    grounding_test["results"]["RESISTANCE_FB"] = cond_val
    grounding_test["results"]["RESISTANCE_PIPES"] = loop_val
    grounding_test["results"]["RESISTANCE_PIPE_GND"] = loop_gnd_val
    #check_for_problems(sheet, grounding_test, [tests.index('COND'), tests.index("WEIGH")])

    #
    # Weight
    petal_weight, weight_pass = get_res_and_accep(sheet, tests.index("WEIGH"))

    #
    # Metrology AVS
    metrology_test = create_metrology_test(session, SN, test_date, operator)
    metrology_test["results"]["LOCATOR1_DIAMETER"] = get_float(sheet['g{}'.format(tests.index("PL01_DIAM"))])
    metrology_test["results"]["LOCATOR2_DIAMETER"] = get_float(sheet['g{}'.format(tests.index("PL02_DIAM"))])
    metrology_test["results"]["LOCATOR3_DIAMETER"] = get_float(sheet['g{}'.format(tests.index("PL03_DIAM"))])
    metrology_test["results"]["LOCATOR2_X"] = get_float(sheet['g{}'.format(find_idx(tests, "PL02_X")[0])])
    metrology_test["results"]["LOCATOR2_Y"] = get_float(sheet['g{}'.format(find_idx(tests, "PL02_Y")[0])])
    metrology_test["results"]["LOCATOR3_X"] = get_float(sheet['g{}'.format(find_idx(tests, "PL03_X")[0])])
    metrology_test["results"]["LOCATOR3_Y"] = get_float(sheet['g{}'.format(find_idx(tests, "PL03_Y")[0])])
    metrology_test["results"]["FIDUCIAL1_DIAMETER"] = get_float(sheet["g{}".format(find_idx(tests, "FD01_DIAM")[0])])
    metrology_test["results"]["FIDUCIAL1_X"] = get_float(sheet["g{}".format(find_idx(tests, "FD01_X")[0])])
    metrology_test["results"]["FIDUCIAL1_Y"] = get_float(sheet["g{}".format(find_idx(tests, "FD01_Y")[0])])
    metrology_test["results"]["FIDUCIAL2_DIAMETER"] = get_float(sheet["g{}".format(find_idx(tests, "FD02_DIAM")[0])])
    metrology_test["results"]["FIDUCIAL2_X"] = get_float(sheet["g{}".format(find_idx(tests, "FD02_X")[0])])
    metrology_test["results"]["FIDUCIAL2_Y"] = get_float(sheet["g{}".format(find_idx(tests, "FD02_Y")[0])])
    metrology_test["results"]["ANGLE_VCHANNEL"] = get_float(sheet["g{}".format(find_idx(tests, "VANGL")[0])])
    metrology_test["results"]["ENVELOPE"] = get_float(sheet["g{}".format(find_idx(tests, "ENVEL")[0])])
    metrology_test["results"]["COPLANARITY_FRONT"] = get_float(sheet["g{}".format(find_idx(tests, "F.PL_PLAN")[0])])
    metrology_test["results"]["LOCAL_FLATNESS_FRONT"] = get_float(sheet["g{}".format(find_idx(tests, "F.FS_PLAN")[0])], '/')
    metrology_test["results"]["PARALLELISM_FRONT"] = get_float(sheet["g{}".format(find_idx(tests, "F.PARAL")[0])])
    metrology_test["results"]["COPLANARITY_BACK"] = get_float(sheet["g{}".format(find_idx(tests, "B.PL_PLAN")[0])])
    metrology_test["results"]["LOCAL_FLATNESS_BACK"] = get_float(sheet["g{}".format(find_idx(tests, "B.FS_PLAN")[0])], '/')
    metrology_test["results"]["PARALLELISM_BACK"] = get_float(sheet["g{}".format(find_idx(tests, "B.PARAL")[0])])

    # Get defects
    check_for_problems(sheet, metrology_test, [tests.index("WEIGH")+1, sheet.max_row])

    return vi_test, delamination_test, grounding_test, metrology_test, batch, petal_weight


def find_label(sheet, label, column, max_row=20):
    """Find label in given column

    Args:
        sheet (): The spread sheet
        label (): The label to search for
        column (): The column to scan.

    Return:
        indx (int) - the index. <0 means not found.
    """
    indx = -1
    for i in range(1, 15):
        val = sheet["{}{}".format(column, i)].value
        if val is None:
            continue

        if val.find(label)>=0:
            indx = i
            break

    return indx

def readProductionSheet(session, file_path, SN):
    """Read data fro AVS PS.

    Args:
        session: the DB session
        file_path: path of input file
        SN: The serial number
        write_json: if true, test json is writen to file.

    """
    try:
        wb = XL.load_workbook(file_path)
    except InvalidFileException as ee:
        print("Could not open input file: ", file_path)
        print(ee.message)
        return None

    # Assume active sheet is the good one, otherwise will have tofind in wb.sheetnames
    sheet = wb.active
    if sheet.max_row > 30 or sheet.max_column > 8:
        raise AVSDataException("Wrong PS file")

    # Find the start
    indx = find_label(sheet, "PETAL", "A")
    if indx < 0:
        # Try with second column
        indx = find_label(sheet, "PETAL", "B")
        if indx < 0:
            print("Wrong Production Sheet.")
            return None
        else:
            icol = ord('B')
            ccol = 'B'
    else:
        icol = ord('A')
        ccol = 'A'

    i_items = find_label(sheet, "DRAWING", ccol)
    if i_items < 0:
        print("Wrong Production Sheet.")
        return None

    i_items += 1

    # Get the SN of the fron facesheet and create the Petal SN
    ID = sheet["{}{}".format(chr(icol+2), i_items)].value.strip()
    SN = get_coreID(ID)
    nn = get_int(sheet["{}{}".format(chr(icol+1), indx)])
    petal_id = "PPC.{:03d}".format(nn)
    mould_id = sheet["{}{}".format(chr(icol+1), indx+3)].value

    # find the date (use the end date)
    start_date = sheet["{}{}".format(chr(icol+4), indx+2)].value
    end_date = sheet["{}{}".format(chr(icol+4), indx+3)].value
    test_date = start_date

    manager = sheet["{}{}".format(chr(icol+4), indx+1)].value

    # Manufacturing
    id_col = chr(icol+2)
    w_col = chr(icol+3)
    n_col = chr(icol+4)
    comments = get_comments(sheet['a25'].value)
    manufacturing = create_manufacturing(session, SN, test_date, manager, comments=comments)
    manufacturing['properties']['START_DATE'] = ITkDButils.get_db_date(start_date)
    manufacturing['properties']['FINISH_DATE'] = ITkDButils.get_db_date(end_date)
    manufacturing["properties"]["MOULD_ID"] = mould_id
    manufacturing["properties"]["PROCESS_DOCUMENT"] = sheet["{}{}".format(chr(icol+4), indx)].value
    manufacturing["results"]["LOCATOR_A"] = sheet["{}{}".format(id_col, i_items+2)].value
    manufacturing["results"]["LOCATOR_B"] = sheet["{}{}".format(id_col, i_items+3)].value
    manufacturing["results"]["LOCATOR_C"] = sheet["{}{}".format(id_col, i_items+4)].value
    manufacturing["results"]["HONEYCOMBSET"] = split_comp_list( sheet["{}{}".format(id_col, i_items+5)].value)
    manufacturing["results"]["EPOXY_ADHESIVE"] = split_comp_list(sheet["{}{}".format(id_col, i_items+8)].value)
    manufacturing["results"]["EPOXY_PUTTY"] = split_comp_list( sheet["{}{}".format(id_col, i_items+9)].value)
    manufacturing["results"]["EPOXY_CONDUCTIVE"] = split_comp_list( sheet["{}{}".format(id_col, i_items+10)].value)

    # Weighing
    weighing = create_weight(session, SN, test_date, manager)
    scol = "{}{}:{}{}".format(w_col, i_items, w_col, i_items+10)
    comp_weight = [get_float(x[0]) for x in sheet[scol]]
    petal_weight = sum([float(x) for x in comp_weight])
    weighing["results"]["WEIGHT_FACING_FRONT"] = comp_weight[0]
    weighing["results"]["WEIGHT_FACING_BACK"] = comp_weight[1]
    weighing["results"]["WEIGHT_LOCATOR_A"] = comp_weight[2]
    weighing["results"]["WEIGHT_LOCATOR_B"] = comp_weight[3]
    weighing["results"]["WEIGHT_LOCATOR_C"] = comp_weight[4]
    weighing["results"]["WEIGHT_COOLINGLOOPASSEMBLY"] = comp_weight[6]
    weighing["results"]["WEIGHT_HONEYCOMBSET"] = comp_weight[5]
    weighing["results"]["WEIGHT_EPOXYADHESIVE"] = comp_weight[8]
    weighing["results"]["WEIGHT_EPOXYPUTTY"] = comp_weight[9]
    weighing["results"]["WEIGHT_EPOXYCONDUCTIVE"] = comp_weight[10]
    weighing["results"]["WEIGHT_CORE"] = petal_weight

    # Comments
    for i in range(i_items, i_items+11):
        cell_id = sheet['{}{}'.format(ccol, i)].value
        comment = sheet['{}{}'.format(n_col, i)].value
        if comment is not None:
            comment = comment.strip()
            if len(comment):
                msg = "{}: {}".format(cell_id, comment)
                weighing["comments"].append(msg)

    DESY = {
        "FacingFront": sheet["{}{}".format(id_col, i_items)].value.strip(),
        "FacingBack": sheet["{}{}".format(id_col, i_items+1)].value.strip(),
        "CoolingLoop": sheet["{}{}".format(id_col, i_items+6)].value.strip(),
        "AllcompSet": sheet["{}{}".format(id_col, i_items+7)].value.strip(),
        "HoneyCombSet": manufacturing["results"]["HONEYCOMBSET"]
    }

    return manufacturing, weighing, DESY, petal_id


if __name__ == "__main__":
    parser = ArgumentParser()
    parser.add_argument('files', nargs='*', help="Input files")
    parser.add_argument("--SN", dest="SN", type=str, default="SNnnn",
                        help="Module serial number")

    options = parser.parse_args()
    if len(options.files) == 0:
        print("I need an input file")
        sys.exit()

    dlg = ITkDBlogin.ITkDBlogin()
    client = dlg.get_client()
    if client is None:
        print("Could not connect to DB with provided credentials.")
        dlg.die()
        sys.exit()

    fnam = Path(options.files[0]).expanduser().resolve()
#    readProductionSheet(fnam, options.SN)
    readFATfile(client, fnam, options.SN)
    dlg.die()



#!/usr/bin/env python3
"""A collection of utilities for sensor data."""
import io
from pathlib import Path
import numpy as np
import warnings
import tempfile
from itkdb_gtk import ITkDButils

#
#   The following is taken from
#   https://gitlab.cern.ch/atlas-itk/sw/db/production_database_scripts.git
#
mm = 1e-3
cm = 1e-2
UpperV = 500  # For sensor to pass, I must be < Imax up until this voltage and no breakdown must be detected before then.
StabilityV = 700  # Voltage with multiple current readings to check stability

AreaDict = {
    "Unknown": (97.621 - 0.450) * (97.950 - 0.550) * mm * mm,
    #
    "ATLAS12": 95.7 * 95.64 * mm * mm,
    "ATLAS17LS": (97.621 - 0.450) * (97.950 - 0.550) * mm * mm,
    #
    "ATLAS18R0": 89.9031 * cm * cm,
    "ATLAS18R1": 89.0575 * cm * cm,
    "ATLAS18R2": 74.1855 * cm * cm,
    "ATLAS18R3": 80.1679 * cm * cm,
    "ATLAS18R4": 87.4507 * cm * cm,
    "ATLAS18R5": 91.1268 * cm * cm,
    #
    "ATLAS18SS": 93.6269 * cm * cm,
    "ATLAS18LS": 93.6269 * cm * cm,
    #
    "ATLASDUMMY18": (97.621 - 0.450) * (97.950 - 0.550) * mm * mm,
}


NO_BD_CONST = 9.99e99

def LocateMicroDischarge(
    I,
    V,
    sm_window=2,
    bd_limit=5.5,
    allow_running_bd=True,
    use_additional_cond=False,
    tolerence=0.05,
    voltage_span=4,
    fit_window=5,
):
    """
    Function for BDV estimation - if questions please contact Vera Latonova (vera.latonova@cern.ch).
    I,V must have same shape and voltages must be in ascending order,
    same indexes of I&V arrays must correspond each other,
    only invalid data or holdstep should be stripped before
    but it is not necessary. Measurments U=0&I=0 are removed.
    If there is same or higher amount of same voltages in row than
    sm_window, from this sequence we cannot estimete dI/dV and
    we have to remove this averaged point.

    It is assumed that only parameter use_additional_cond would be
    changed by caller. Changing of other parameters may affect
    BDV unexpectedly.


    @param[in] I                   - array of currents without any cut
    @param[in] V                   - array of voltages, ascending order, without any cut
    @param[in] sm_window           - size of smoothing window
    @param[in] bd_limit            - BD limit for |U| < 500V
    @param[in] allow_running_bd    - allow increase bd_limit for |U| > 500
    @param[in] use_additional_cond - use additional BD contition
    @param[in] tolerence           - configuration of additional condition
    @param[in] voltage_span        - max width of hump on spectra which may be neglected
                                     in voltage steps in additional contition
    @param[in] fit_window          - number of points used for linear fit before BD voltage

    @return BD voltage (always positive) or NO_BD_CONST = 9.99e99 if not found.
    """

    # add nan to the end of array
    V = np.abs(V)
    I = np.abs(I)

    # skip zeros
    ind = np.where(np.logical_or(I != 0, V != 0))
    V = V[ind]
    I = I[ind]

    V_ = np.append(V, np.nan * np.ones(sm_window - 1))
    I_ = np.append(I, np.nan * np.ones(sm_window - 1))

    # make 2D array of I's, V's each row_ind shifted by row_ind index
    # i.e from array [1,3,5] we make (for sm_window=2) 2D array
    # [  1,3,5,nan]
    # [nan,5,1,3]
    # than get average from each column -> I_avg, V_avg
    r = np.arange(sm_window)

    V2 = np.outer(np.ones(sm_window), V_)
    row_ind, col_ind = np.ogrid[: V2.shape[0], : V2.shape[1]]
    col_ind = col_ind - r[:, np.newaxis]
    V2 = V2[row_ind, col_ind]
    # strip fields with nans
    V2 = np.transpose(V2[:, (sm_window - 1) : -(sm_window - 1)])

    I2 = np.outer(np.ones(sm_window), I_)
    row_ind, col_ind = np.ogrid[: I2.shape[0], : I2.shape[1]]
    col_ind = col_ind - r[:, np.newaxis]
    I2 = I2[row_ind, col_ind]
    I2 = np.transpose(I2[:, (sm_window - 1) : -(sm_window - 1)])

    # get V & I averages
    try:
        V_avg = np.average(V2, axis=1)
        I_avg = np.average(I2, axis=1)
    except ZeroDivisionError:
        # not enough data
        return NO_BD_CONST

    # find dI / dV array
    # I'm not able to write this without cycle
    dIdV = np.array([])
    for i in range(V2.shape[0]):
        with warnings.catch_warnings():
            warnings.filterwarnings("error")
            try:
                dIdV = np.append(dIdV, np.polyfit(V2[i, :], I2[i, :], 1)[0])
            except (np.RankWarning, TypeError):
                dIdV = np.append(dIdV, np.nan)

    # stripping U[n] == U[n+1] (i.e. hodlsetp) => fit cannot be sucessful =>
    # dIdV is nan @holdstep
    ind = np.where(np.isfinite(dIdV))
    I_avg = I_avg[ind]
    V_avg = V_avg[ind]
    dIdV = dIdV[ind]

    # get running BDV limit & compare
    bd_limit_running = bd_limit + np.where(
        allow_running_bd and V_avg > 500, (V_avg - 500.0) / 100.0, 0
    )
    V_avg_BD_ind = dIdV / (I_avg / V_avg) > bd_limit_running
    V_avg_BD = V_avg[V_avg_BD_ind]

    # Estimate BDV
    BDV = np.array([])

    # no break-down
    if V_avg_BD.shape == (0,):
        return NO_BD_CONST

    # if V_avg_BD_ind[0] == True ... BDV <- V[0]
    # for others V_avg_BD_ind[n] == True BDV <- (V_avg[n] + V_avg[n-1])/2
    if V_avg_BD_ind[0]:
        BDV = np.append(BDV, V[0])
    V_avg_BD_ind[0] = False

    BDV = np.append(
        BDV,
        (V_avg[np.where(V_avg_BD_ind)] + V_avg[np.where(V_avg_BD_ind)[0] - 1]) / 2.0,
    )

    ###########################################################################
    ## Application of additional condition ####################################
    ###########################################################################
    if not use_additional_cond:
        return BDV[0]

    # get index if V <= BDV
    B = np.where(np.less.outer(BDV, V))
    col_ind = np.mgrid[: BDV.shape[0], : V.shape[0]][1]
    col_ind[B[0], B[1]] = 0
    V_BDV_ind = np.max(col_ind, axis=1)

    back_ok_v_ind = 0
    while True:
        with warnings.catch_warnings():
            warnings.filterwarnings("error")
            try:
                a, b = np.polyfit(
                    V[
                        max(back_ok_v_ind, V_BDV_ind[0] - fit_window) : max(
                            back_ok_v_ind, V_BDV_ind[0]
                        )
                    ],
                    I[
                        max(back_ok_v_ind, V_BDV_ind[0] - fit_window) : max(
                            back_ok_v_ind, V_BDV_ind[0]
                        )
                    ],
                    1,
                )
            except (np.RankWarning, TypeError):
                return BDV[0]

        ind = np.where(1 - (a * V + b) / I <= tolerence)[0]
        try:
            back_ok_v_ind = np.min(ind[ind > V_BDV_ind[0] + 1])
        except ValueError:
            # sensor is not going back
            return BDV[0]
        # hump is too long -- it cannot be skipped
        if back_ok_v_ind - V_BDV_ind[0] > voltage_span:
            return BDV[0]

        # skip BDVs inside hump
        ind = BDV >= V[back_ok_v_ind]
        BDV = BDV[ind]
        V_BDV_ind = V_BDV_ind[ind]
        if V_avg_BD.shape == (0,):
            return NO_BD_CONST
        
    return NO_BD_CONST


def scale_iv(I, T1, T2):
    """Normalize corrent  to given temperature (T2)

    Args:
        I (array): Current
        T1 (float): Original temperature
        T2 (float): New temperature.

    Return:
        Array with scaled currents.

    """
    factor = (T2 / T1) ** 2 * np.exp((-1.21 / 8.62) * (1 / T2 - 1 / T1))
    return factor * I


def sensor_data_to_json(session, mdata, mod_type, logger=None):
    """Read a Sensor data file and return the corresponding JSon for hte PDB.

    Args:
        mdata: Data as returned by :function:`read_sensor_file`.
        logger: an object that writes messages via a write_message method
    """
    # Check if we are dealing with sensors or modules
    is_module = "Module_SN" in mdata

    if is_module:
        test = ITkDButils.get_test_skeleton(session, "MODULE", mdata["TestType"])
        tp = "ATLAS18{}".format(mod_type[0:2])
        area = AreaDict[tp] / cm**2
        SN = mdata["Module_SN"]

    else:
        test = ITkDButils.get_test_skeleton(session, "SENSOR", mdata["TestType"])
        area = AreaDict[mod_type] / cm**2
        SN = mdata["Component"]

    if logger:
        logger.write_message("Analyzing {}\n".format(SN))

    # The data arrays
    V = np.abs(mdata["curve"]["V"])
    I = np.abs(mdata["curve"]["I"])
    passed = True

    # Find Current @ 500V
    indx = np.where(V == 500)[0]
    i_500 = I[indx][0] / area
    if logger:
        logger.write_message("I @ 500V = {:.2f} nA/cm2\n".format(i_500))

    # Compute current stability
    IStability = abs(I[abs(V) == StabilityV])
    IVariation = -1
    if np.size(IStability) > 1:  # Maybe make == 4?
        IVariation = abs(np.std(IStability) / np.mean(IStability))

    if logger:
        logger.write_message("I stability = {:.6f} nA\n".format(IVariation))

    # Search for Micro discharges
    # Check for micro-discharge in non-normalized current,
    # removing duplicate Voltage entries (e.g. for stability measurements)
    comments = []
    defects = []
    UniqueVs, UniqueIndices = np.unique(V, return_index=True)
    MicroDischargeV = LocateMicroDischarge(I[UniqueIndices], UniqueVs)
    if MicroDischargeV < np.max(V):
        comments.append("Found micro discharge: {:.1f} V\n".format(MicroDischargeV))
        if logger:
            logger.write_message(comments[-1])

        if MicroDischargeV < UpperV:
            txt = "microdischarge happening before {:.1f}V.".format(UpperV)
            defects.append({
                        "name": "MicroDischarge",
                        "description": txt,
                        "properties": {}
                    }
            )
            if logger:
                logger.write_message("...{}. FAILED\n".format(txt))
                
            passed = False
    else:
        if MicroDischargeV == NO_BD_CONST:
            MicroDischargeV = 700.0

    test["component"] = SN
    test["institution"] = mdata["Institute"]
    test["runNumber"] = mdata["RunNumber"]
    test["date"] = ITkDButils.get_db_date(
        "{} {}".format(mdata["Date"], mdata["Time"])
    )
    test["passed"] = passed
    test["problems"] = False
    test["properties"]["VBIAS_SMU"] = mdata["Vbias_SMU"]
    test["properties"]["RSERIES"] = mdata["Rseries"]
    test["properties"]["TEST_DMM"] = mdata["Test_DMM"]
    test["properties"]["RSHUNT"] = mdata["Rshunt"]
    test["properties"]["RUNNUMBER"] = mdata["RunNumber"]
    test["properties"]["COMMENTS"] = mdata["Comments"]
    test["properties"]["ALGORITHM_VERSION"] = "0.0.0"
    if is_module:
        test["properties"]["SOFTWARE_TYPE_VERSION"] = "pyProbe"
        test["properties"]["MODULE_STAGE"] = mdata["Module_Stage"]

    test["results"]["TEMPERATURE"] = mdata["Temperature"]
    test["results"]["HUMIDITY"] = mdata["Humidity"]
    test["results"]["VBD"] = MicroDischargeV
    test["results"]["I_500V"] = i_500
    test["results"]["VOLTAGE"] = -np.abs(V)
    test["results"]["CURRENT"] = -np.abs(I)
    test["results"]["RMS_STABILITY"] = IVariation
    test["results"]["SHUNT_VOLTAGE"] = np.zeros(V.shape)
    test["defects"] = defects
    test["comments"] = comments

    return test


def read_sensor_file(fnam):
    """Read a data file. Return dictionary with all teh data."""
    labels = []
    metadata = {}
    with open(fnam, "r", encoding="utf-8") as ifile:
        first = True
        for line in ifile:
            if first:
                first = False
                ipos = line.rfind(".")
                metadata["fname"] = line[:ipos]
                continue

            if line.find("Voltage [V]") >= 0 or line.find("Voltage[V]") >= 0:
                labels = line.split("\t")
                break

            rc = line.find(":")
            if rc >= 0:
                key = line[:rc].strip()
                val = line[rc + 1 :].strip()
                if key in ["Temperature", "Humidity"]:
                    metadata[key] = float(val)
                else:
                    metadata[key] = val

        V = []
        I = []
        S = []
        for line in ifile:
            data = [float(s) for s in line.split()]
            V.append(data[0])
            I.append(data[1])
            try:
                S.append(data[2])
            except IndexError:
                S.append(0.0)

        metadata["curve"] = {
            "V": np.abs(np.array(V)),
            "I": np.abs(np.array(I)),
            "S": np.abs(np.array(S)),
            "labels": labels,
        }
        return metadata


def save_sensor_data(fnam, mdata, name=None):
    """Save sensor dat in file with the proper format.

    Args:
        fnam: file name or file object.
        mdata (dict): data as returned by :function:`read_sensor_file`
        
    """
    if hasattr(fnam, "write") and callable(fnam.write):
        data_out = fnam
        if name:
            fnam = name
        else:
            fnam = build_file_name(mdata)
            
    else:
        data_out = open(fnam, 'w', encoding="utf-8")
        fnam = Path(fnam).name
        
    is_module = "Module_SN" in mdata
    if is_module:
        SN = mdata["Module_SN"]
    else:
        SN = mdata["Component"]
        
    if is_module:
        items = [
            "Type",
            "Wafer",
            "Module_SN",
            "Module_Stage",
            "Date",
            "Time",
            "Institute",
            "TestType",
            "Vbias_SMU",
            "Rseries",
            "Test_DMM",
            "Rshunt",
            "Software type and version, fw version",
            "RunNumber",
            "Temperature",
            "Humidity",
            "Comments",
        ]
    else:
        items = [
            "Type",
            "Batch",
            "Wafer",
            "Component",
            "Date",
            "Time",
            "Institute",
            "TestType",
            "Vbias_SMU",
            "Rseries",
            "Test_DMM",
            "Rshunt",
            "RunNumber",
            "Temperature",
            "Humidity",
            "Comments",
        ]
    
    data_out.write("{}\n".format(fnam))
    for key in items:
        if key == "Module_SN" or key == "Component":
            data_out.write("{}: {}\n".format(key, SN))
        else:
            data_out.write("{}: {}\n".format(key, mdata[key]))

    labels = [lbl for lbl in mdata["curve"]["labels"]]
    if len(labels) < 3:
        labels.append("Shunt_voltage [mV]")

    for il, label in enumerate(labels):
        if il:
            data_out.write("\t")
        data_out.write(label)
    data_out.write("\n")


    # The data arrays
    V = np.abs(mdata["curve"]["V"])
    I = np.abs(mdata["curve"]["I"])
    S = np.abs(mdata["curve"]["S"])
    
    ndata = len(V)
    for i in range(ndata):
        data_out.write("{:10.2f}\t{:10.2f}\t{:10.2f}\n".format(V[i], I[i], S[i]))

    print(data_out.name)
    data_out.close()

def build_file_name(mdata):
    """Create a file name from the data."""
    is_module = "Module_SN" in mdata
    if is_module:
        SN = mdata["Module_SN"]
        fnam = "{}_{}_IV_{}".format(SN, mdata["Module_Stage"], mdata["RunNumber"])
    else:
        SN = mdata["Component"]
        fnam = "{}-W{}_IV_{}".format(mdata["Batch"], mdata["Wafer"], mdata["RunNumber"])

    return fnam


"""Utilities for the inteaction with the ITkDB."""
import mimetypes
from collections.abc import Iterable
from datetime import datetime
from pathlib import Path

import dateutil.parser

# The response of the DB
db_response = ""


# define an Attachment object.
class Attachment(object):
    """Encapsulates Attachment information."""

    def __init__(self, path=None, title=None, desc=None):
        """Initialization."""
        if path is not None:
            self.path = Path(path).expanduser().resolve()
        else:
            self.path = None

        self.title = title
        self.desc = desc


def is_iterable(obj):
    """Tell if an object is iterable. Strings are not considered iterables."""
    if isinstance(obj, Iterable):
        if isinstance(obj, str) or isinstance(obj, bytes):
            return False
        else:
            return True
    else:
        return False


def get_db_response():
    """Return the DB response.

    It is stores in a global variable. Trust the function if call
    right after your interaction with the DB.
    """
    global db_response
    return db_response


def get_db_date(timestamp=None):
    """Convert a date string into the expected DB format.

    Args:
        timestamp: A date in string format

    """
    def date2string(the_date):
        out = the_date.isoformat(timespec='milliseconds')
        if out[-1] not in ['zZ']:
            out += 'Z'

        return out
        # return the_date.strftime("%Y-%m-%dT%H:%M:%S.000Z")

    out = None
    if timestamp is None:
        out = date2string(datetime.now())
    elif isinstance(timestamp, datetime):
        out = date2string(timestamp)
    else:
        try:
            this_date = dateutil.parser.parse(timestamp)
            out = date2string(this_date)
        except Exception:
            out = ""

    return out


def registerPetalCore(client, SN, alias, HC_id=None):
    """Register a Petal Core in the DB.

    Args:
        SN: Serial Number
        alias: The alias of the Petal
        HC_id: Comma separated list of HC identifiers.

    Returns:
        _type_: _description_

    """
    global db_response
    dto = {
        "project": 'S',
        "subproject": "SE",
        "institution": "AVS",
        "componentType": "CORE_PETAL",
        "type": "CORE_AVS",
        "serialNumber": SN,
        "properties": {'DESY_ID': alias}
    }
    if HC_id is not None:
        dto["properties"]["HC_ID"] = HC_id

    db_response = client.post('registerComponent', json=dto)
    try:
        return db_response['component']

    except KeyError:
        return None


def create_component_attachment(client, SN, file_path, title=None, description="", item_type="component"):
    """Create an attachment to the given component.

    Args:
        client: The DB client
        SN: The SN of the component.
        file_path: The pat to th efile to be attached.
        title: title of the file
        description: Description of the attachment
        item_type: the type of DB object to attach the information

    """
    global db_response
    db_response = None
    path = Path(file_path).expanduser().resolve()
    filetype = mimetypes.guess_type(path)
    if title is None:
        title = path.name

    if item_type == "component":
        c_item = "component"
        cmmd = "createComponentAttachment"

    elif item_type == "shipment":
        c_item = "shipment"
        cmmd = "createShipmentAttachment"

    else:
        print("Unknown dB object")
        db_response = None
        return

    data = {
        c_item: SN,
        'title': title,
        'description': description,
        'type': 'file',
        'url': path.as_uri(),
    }
    # Get attachment data
    attachment = {'data': (path.name, open(path.as_posix(), 'rb'), filetype)}
    try:
        db_response = client.post(cmmd, data=data, files=attachment)

    except Exception as e:
        print("Error creating attachment\n", e)

    return db_response


def set_component_property(client, SN, property, value):
    """Set the value of an object property.

    Args:
        client: The DB client
        SN: The object SN
        property: The property name
        value: The property value

    """
    global db_response
    try:
        db_response = client.post('setComponentProperty',
                                  json={'component': SN,
                                        'code': property,
                                        'value': value})
        return db_response

    except Exception as e:
        print("Problems setting {} to {} in {}:\n\t{}".format(
            property, value, SN, str(e)))
        return None


def assemble_component(client, parent, child):
    """Assemble child into parent.

    Args:
        client: The DB client
        parent: The parent object or container.
        child: The child to be assembled.

    """
    global db_response
    try:
        db_response = client.post("assembleComponent",
                                  json={'parent': parent, 'child': child})
        return db_response

    except Exception as e:
        print("Problems assemblying {} into {}:\n\t{}".format(
            child, parent, str(e)))
        return None


def set_object_stage(client, SN, stage):
    """Set stage of object

    Args:
        client: DB session
        SN: Serial number
        stage: Stage

    """
    global db_response
    try:
        db_response = client.post("setComponentStage",
                                  json={"component": SN, "stage": stage})
        return db_response

    except Exception as e:
        print("Problems changing stage of {} into {}:\n\t{}".format(
            SN, stage, str(e)))
        return None


def get_DB_component(client, SN):
    """Get ta component by its serial number."""
    global db_response

    try:
        db_response = client.get("getComponent", json={"component": SN})
        return db_response

    except Exception as e:
        db_response = str(e)

    try:
        out = client.get('getComponent', json={'component': SN, "alternativeIdentifier":True})
        db_response = out
        return out

    except Exception as e:
        db_response = str(e)
        return None

def upload_test(client, data, attachments=None):
    """Upload a test to the DB.

    Args:
        client: The DB client
        data (dict): A dictionary with all the elements of thee test.
        attachments ([Attachment]): one or more (in a list) Attachment to the test

    Return:
        resp: The response of the DB session.

    """
    global db_response
    try:
        db_response = client.post("uploadTestRunResults", json=data)
        testRun = db_response["testRun"]["id"]
        if attachments is not None:
            if not isinstance(attachments, Iterable):
                attachments = (attachments)

            for att in attachments:
                path = Path(att.path).expanduser().resolve()
                if not path.exists():
                    print("File {} does not exist".format(path))
                    continue

                data = {"testRun": testRun,
                        "title": att.title if att.title is not None else path.name,
                        "description": att.desc if att.desc is not None else path.name,
                        "type": "file",
                        }
                filetype = mimetypes.guess_type(path.name)
                attachment = {'data': (path.name, open(path.as_posix(), 'rb'), filetype[0])}
                db_response = client.post('createTestRunAttachment',
                                          data=data,
                                          files=attachment)

        return None

    except Exception as e:
        return (str(e))


def create_shipment(session, sender, recipient, items, name=None, send=False,
                    type="domestic", attachment=None, comments=None):
    """Create a chipment.

    Args:
        session : The itkdb session
        sender : The sender ID
        recipient : The recipient ID
        items : A list of SN of the items to be shipped.
        name: the name of the shipment.
        send: If true, the status of the shipment is updated to inTransit
        type (optional): Type of shipment. Defaults to "domestic".
        attachment (optional, :class:`Attachment`): :class:`Attachment` object.
        comments (optional): comments for the shipment

    """
    global db_response
    if name is None:
        name = "From {} to {}".format(sender, recipient)

    if type not in ["domestic", "intraContinental", "continental"]:
        db_response = "Wrong shipment type."
        return None

    if len(items) == 0:
        db_response = "Empty lit of items"
        return None

    the_comments = None
    if comments is not None:
        if not is_iterable(comments):
            the_comments = [comments]
        else:
            the_comments = comments

    data = {
        "name": name,
        "sender": sender,
        "recipient": recipient,
        "type": type,
        "shipmentItems": items,
    }

    if the_comments:
        data["comments"] = the_comments

    try:
        db_response = session.post("createShipment", json=data)

    except Exception as E:
        db_response = str(E)
        return None

    if not send:
        return db_response

    shipment_id = db_response["id"]
    payload = {
        "shipment": shipment_id,
        "status": "inTransit",
    }

    # The items
    items = []
    for item in db_response["shipmentItems"]:
        the_item = {"code": item['code'], "delivered": True}
        items.append(the_item)

    payload["shipmentItems"] = items

    db_response = None
    try:
        db_response = session.post("setShipmentStatus", json=payload)

    except Exception as E:
        db_response = str(E)
        return None

    if attachment:
        try:
            rc = create_component_attachment(session, shipment_id,
                                             attachment.path,
                                             attachment.title,
                                             attachment.desc,
                                             "shipment")
            db_response = rc

        except Exception:
            db_response = "Could not add the attachment."

    return db_response


def set_shipment_status(client, data):
    """Update shipment status."""
    global db_response
    try:
        db_response = client.post("setShipmentStatus", json=data)
        return db_response

    except Exception:
        return None


def from_full_test_to_test_data(full_test):
    """Conver getTest output to json needed to upload the test."""
    test = {
        "component": None,
        "testType": full_test["testType"]["code"],
        "institution": full_test["institution"]["code"],
        "runNumber": full_test["runNumber"],
        "date": full_test["date"],
        "passed": full_test["passed"],
        "problems": full_test["problems"],
        "properties": {},
        "results": {},
        "comments": [],
        "defects": []
    }

    try:
        test["component"] = full_test["components"][0]["serialNumber"]

    except Exception:
        print("from_full_test_to_test_data\nPossible error: no SN found for component.")
        pass

    for P in full_test["properties"]:
        test["properties"][P["code"]] = P['value']

    if full_test["results"]:
        for P in full_test["results"]:
            test["results"][P["code"]] = P['value']

    if full_test["comments"]:
        for C in full_test["comments"]:
            test["comments"].append(C["comment"])

    if full_test["defects"]:
        for D in full_test["defects"]:
            test["defects"].append({"name": D["name"],
                                    "description": D["description"],
                                    "properties": D["properties"]})

    return test


def get_testrun(session, test_id, out_type="object"):
    """Retrieve information about a given test.

    Args:
        session : The itkdb session
        test_id : The ID of the test.
        out_type (optional): Type of output (full or object). Defaults to "object".

    """
    global db_response
    try:
        db_response = session.get("getTestRun", json={"testRun": test_id})
        return db_response

    except Exception:
        return None


def get_test_skeleton(session, component, test_code, userdef={}, uservalues={}):
    """Get the skeleton of the given test.

    Args:
        session: The DB session
        component: The component which is tested
        test_code: The test code
        userdef: default values of test parameters, propertines and results.
        uservalues: default values for different types.

    """
    global db_response
    defvalues = {
        "string": "",
        "integer": -9999,
        "float": -9999.0,
        "boolean": True,
    }
    for ky, val in defvalues.items():
        if ky not in uservalues:
            uservalues[ky] = val

    data = {"project": "S", "componentType": component}
    out = session.get("listTestTypes", json=data)
    db_response = out
    the_test = None
    for tst in out:
        # print(tst['code'])
        if tst['code'] == test_code:
            the_test = tst
            break

    if the_test is None:
        print("test {} not found for {}".format(test_code, component))
        return None

    # for prop in the_test["properties"]:
    #     print("{}; {}".format(prop["code"], prop["name"]))

    # for prop in the_test["parameters"]:
    #     print("{}; {}".format(prop["code"], prop["name"]))

    skltn = {
        "component": None,
        "testType": test_code,
        "institution": None,
        "runNumber": "-1",
        "date": get_db_date(),
        "passed": True,
        "problems": False,
        "properties": {},
        "results": {},
        "comments": [],
        "defects": []
    }

    # Set default values
    for key, val in skltn.items():
        if key in userdef:
            if isinstance(val, dict):
                continue

            val = userdef[key]

    def get_default(vin, default=None):
        # print(json.dumps(vin, indent=1))
        if vin['valueType'] == "single":
            vtype = vin['dataType']
            val = None
            if vtype in uservalues:
                val = uservalues[vtype]
            else:
                # print("default for data type ", vtype, " not found")
                val = None

        else:
            val = None

        return val

    # SEt the properties
    for prop in the_test['properties']:
        key = prop["code"]
        if 'properties' in userdef and key in userdef['properties']:
            skltn['properties'][key] = userdef['properties'][key]
        else:
            skltn['properties'][key] = get_default(prop)

    # SEt the parameters
    for par in the_test['parameters']:
        key = par["code"]
        if 'results' in userdef and key in userdef['results']:
            skltn['results'][key] = userdef['results'][key]
        else:
            skltn['results'][key] = get_default(par)

    return skltn



#!/usr/bin/env python3
"""PB/Hybrid panel Visual inspection GUI.."""
import json
import sys

try:
    import itkdb_gtk

except ImportError:
    from pathlib import Path
    cwd = Path(__file__).parent.parent
    sys.path.append(cwd.as_posix())

from itkdb_gtk import dbGtkUtils, ITkDBlogin, ITkDButils

import gi
gi.require_version("Gtk", "3.0")
from gi.repository import Gtk, Gdk, Gio

HELP_LINK="https://itkdb-gtk.docs.cern.ch"

class PanelVisualInspection(dbGtkUtils.ITkDBWindow):
    """PB/Hybryd panel visual inspection GUI."""
    SN, PASSED, ALL = range(3)

    def __init__(self, session, help=HELP_LINK):
        super().__init__(title="ITkDB Dashboard",
                         session=session,
                         show_search="Find object with given SN.",
                         help=help)

        # action button in header
        button = Gtk.Button()
        icon = Gio.ThemedIcon(name="document-send-symbolic")
        image = Gtk.Image.new_from_gicon(icon, Gtk.IconSize.BUTTON)
        button.add(image)
        button.set_tooltip_text("Click to upload ALL tests.")
        button.connect("clicked", self.upload_tests)
        self.hb.pack_end(button)

        grid = Gtk.Grid(column_spacing=5, row_spacing=1)
        self.mainBox.pack_start(grid, False, False, 5)

        lbl = Gtk.Label(label="Serial Number")
        lbl.set_xalign(0)
        grid.attach(lbl, 0, 0, 1, 1)

        self.SN = dbGtkUtils.TextEntry()
        self.SN.connect("text_changed", self.SN_ready)
        self.SN.widget.set_tooltip_text("Enter SN of PWD or Hybrid panel.")
        grid.attach(self.SN.widget, 1, 0, 1, 1)

        self.panel_type = Gtk.Label(label="")
        grid.attach(self.panel_type, 2, 0, 1, 1)

        # Paned object
        paned = Gtk.Paned(orientation=Gtk.Orientation.VERTICAL)
        paned.set_size_request(-1, 200)
        self.mainBox.pack_start(paned, True, True, 5)

        # the list of attachments
        tree_view = self.create_tree_view()
        paned.add1(tree_view)

        # The text view
        paned.add2(self.message_panel.frame)


        self.show_all()

        dbGtkUtils.setup_scanner(self.get_qrcode)

    def create_tree_view(self, size=150):
        """Create the TreeView with the children."""
        model = Gtk.ListStore(str, bool)
        self.tree = Gtk.TreeView(model=model)

        scrolled = Gtk.ScrolledWindow()
        scrolled.set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.AUTOMATIC)
        scrolled.add(self.tree)
        scrolled.set_size_request(-1, size)
        
        renderer = Gtk.CellRendererText()
        column = Gtk.TreeViewColumn("SN", renderer, text=PanelVisualInspection.SN)
        self.tree.append_column(column)

        renderer = Gtk.CellRendererToggle()
        renderer.set_property("activatable", True)
        renderer.set_property("radio", True)
        renderer.set_padding(5, 0)
        
        x, y = renderer.get_alignment()
        renderer.set_alignment(0, y)
        # renderer.set_property("inconsistent", True)
        renderer.connect("toggled", self.btn_toggled)
        
        column = Gtk.TreeViewColumn("Passed", renderer, active=PanelVisualInspection.PASSED)
        self.tree.append_column(column)

        return scrolled

    def btn_toggled(self, renderer, path, *args):
        """Toggled."""
        model = self.tree.get_model()
        val = not model[path][PanelVisualInspection.PASSED]
        model[path][PanelVisualInspection.PASSED] = val
        

    def button_pressed(self, tree, event):
        """Button pressed on tree view."""
        # double click shows attachments
        if event.button == 1 and event.type == Gdk.EventType.DOUBLE_BUTTON_PRESS:
            select = self.tree.get_selection()
            model, iter = select.get_selected()
            if not iter:
                return

            self.on_show_json(None, (model, iter, model[iter]))
            # self.on_show_attachments(None, (model, iter, model[iter]))
            return

        if event.button != 3:
            return

        # Create popup menu
        select = self.tree.get_selection()
        model, iter = select.get_selected()
        values = None
        if iter:
            values = model[iter]

        if not iter:
            P = tree.get_path_at_pos(event.x, event.y)
            if P:
                print(P[0].to_string())
                iter = model.get_iter(P[0])
                values = model[iter]

        if not values:
            return


    def SN_ready(self, *args):
        """SN is ready in the TextEnttry."""
        SN = self.SN.get_text()
        if not SN.startswith("20U") or len(SN)!=14:
            dbGtkUtils.complain("Invalid Serial Number",
                                    "{}: wrong SN".format(SN))
            return

        if "USED" in SN:
            # Powerboard Carrier
            if not SN[6].isdigit():
                dbGtkUtils.complain("Not a Powerboard Carrier",
                                    "{}: wrong SN for a powerboard carrier".format(SN))
                self.SN.widget.set_text("")
                return

            self.panel_type.set_text("PWB carrier")

        elif "USET" in SN:
            # Hybrid test panel
            if not SN[6].isdigit or int(SN[6])>5:
                dbGtkUtils.complain("Not a Hybrid Test Panel",
                                    "{}: wrong SN for a hybrid test panel".format(SN))
                self.SN.widget.set_text("")
                return

        else:
            dbGtkUtils.complain("Invalid SN.",
                "{}\nNot a PWB carrier not HYB test panel.".format(SN))
            self.SN.widget.set_text("")
            return

        # GEt children.
        panel = ITkDButils.get_DB_component(self.session, SN)
        if panel is None:
            self.write_message(ITkDButils.get_db_response())
            return

        children = []
        for child in panel["children"]:
            if child["component"] is not None:
                if child["componentType"]["name"] == "Powerboard":
                    children.append(child["component"]["serialNumber"])

        model = Gtk.ListStore(str, bool)
        for child in children:
            model.append([SN, True])
            
        self.tree.set_model(model)


    def upload_tests(self, *args):
        """Upload the current test."""
        SN = self.SN.get_text()

    def get_qrcode(self, fd, state, reader):
        """Read SN from scanner."""
        txt = dbGtkUtils.scanner_get_line(reader)
        self.write_message("SN: {}\n".format(txt))


def main():
    """Main entry."""
    # DB login
    dlg = ITkDBlogin.ITkDBlogin()
    client = dlg.get_client()
    if client is None:
        print("Could not connect to DB with provided credentials.")
        dlg.die()
        sys.exit()

    client.user_gui = dlg

    gTest = PanelVisualInspection(client)

    gTest.present()
    gTest.connect("destroy", Gtk.main_quit)
    try:
        Gtk.main()

    except KeyboardInterrupt:
        print("Arrrgggg!!!")

    dlg.die()


if __name__ == "__main__":
    main()